from __future__ import annotations

import numbers
from abc import ABC
from collections.abc import Sequence
from dataclasses import dataclass
from typing import (
    Any,
    ClassVar,
    Generic,
    Literal,
    NoReturn,
    Protocol,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import Self

from cognite.client._constants import MAX_VALID_INTERNAL_ID
from cognite.client.utils._auxiliary import exactly_one_is_not_none, split_into_chunks
from cognite.client.utils.useful_types import SequenceNotStr


@dataclass(frozen=True)
class InstanceId:
    _instance_type: ClassVar[Literal["node", "edge"]]
    space: str
    external_id: str

    def dump(self, camel_case: bool = True, include_instance_type: bool = True) -> dict[str, str]:
        return {"space": self.space, "externalId" if camel_case else "external_id": self.external_id}

    @overload
    @classmethod
    def load_if(cls, data: None) -> None: ...

    @overload
    @classmethod
    def load_if(cls, data: dict[str, str] | tuple[str, str] | Self) -> Self: ...

    @classmethod
    def load_if(cls, data: dict[str, str] | tuple[str, str] | Self | None) -> Self | None:
        # Note: For experimentation - I'd like to add this as a universal method to all classes to avoid
        # the endless spam of 'MyClass.load(foo["bar"]) if "bar" in foo else None' in the codebase!
        if data is None:
            return None
        return cls.load(data)

    @classmethod
    def load(cls, data: dict[str, str] | tuple[str, str] | Self) -> Self:
        if isinstance(data, cls):
            return data
        elif isinstance(data, tuple):
            if len(data) == 2:
                return cls(*data)
            raise ValueError(
                f"Cannot load {data} into {cls}, expected exactly two elements in the tuple, got {len(data)}"
            )

        elif isinstance(data, dict):
            if "externalId" in data:
                return cls(space=data["space"], external_id=data["externalId"])
            if "external_id" in data:
                return cls(space=data["space"], external_id=data["external_id"])
        raise KeyError(f"Cannot load {data} into {cls}, missing 'externalId' or 'external_id' key")

    @property
    def instance_type(self) -> Literal["node", "edge"]:
        return self._instance_type

    def as_tuple(self) -> tuple[str, str]:
        return self.space, self.external_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(space={self.space!r}, external_id={self.external_id!r})"


T_InstanceId = TypeVar("T_InstanceId", bound=InstanceId)

T_ID = TypeVar("T_ID", int, str, InstanceId)


class IdentifierCore(Protocol):
    def as_dict(self, camel_case: bool = True) -> dict: ...

    def as_primitive(self) -> str | int: ...


class Identifier(Generic[T_ID]):
    def __init__(self, value: T_ID) -> None:
        if not isinstance(value, (int, str, InstanceId)):
            raise TypeError(f"Expected id/external_id to be of type int or str, got {value} of type {type(value)}")
        self.__value: T_ID = value

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Identifier):
            return self.__value == other.__value
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.__value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__value!r} [{self.name(camel_case=False)}])"

    @classmethod
    def of_either(
        cls, id_: int | None, external_id: str | None, instance_id: InstanceId | tuple[str, str] | None = None
    ) -> Identifier:
        if id_ is external_id is instance_id is None:
            raise ValueError("Exactly one of id, external id, or instance_id must be specified, got neither")
        elif id_ is not None:
            if external_id is not None or instance_id is not None:
                raise ValueError("Exactly one of id, external id, or instance_id must be specified, got multiple")
            elif not isinstance(id_, int):
                raise TypeError(f"Invalid id, expected int, got {type(id_)}")
            elif not 1 <= id_ <= MAX_VALID_INTERNAL_ID:
                raise ValueError(f"Invalid id, must satisfy: 1 <= id <= {MAX_VALID_INTERNAL_ID}")
        elif external_id is not None:
            if instance_id is not None:
                raise ValueError("Exactly one of id, external id, or instance_id must be specified, got multiple")
            elif not isinstance(external_id, str):
                raise TypeError(f"Invalid external_id, expected str, got {type(external_id)}")
        elif instance_id is not None:
            if isinstance(instance_id, dict | tuple):
                instance_id = InstanceId.load(instance_id)
            if not isinstance(instance_id, InstanceId):
                raise TypeError(f"Invalid instance_id, expected InstanceId, got {type(instance_id)}")
        return Identifier(next(idx for idx in (id_, external_id, instance_id) if idx is not None))

    @classmethod
    def load(
        cls, id: int | None = None, external_id: str | None = None, instance_id: InstanceId | None = None
    ) -> Identifier:
        if id is not None:
            return Identifier(id)
        if external_id is not None:
            return Identifier(external_id)
        if instance_id is not None:
            return Identifier(instance_id)
        raise ValueError("At least one of id, external id, instance_id must be specified")

    def name(self, camel_case: bool = False) -> str:
        if self.is_id:
            return "id"
        elif self.is_instance_id:
            return "instanceId" if camel_case else "instance_id"
        else:
            return "externalId" if camel_case else "external_id"

    def as_primitive(self) -> T_ID:
        return self.__value

    @property
    def is_id(self) -> bool:
        return isinstance(self.__value, int)

    @property
    def is_external_id(self) -> bool:
        return isinstance(self.__value, str)

    @property
    def is_instance_id(self) -> bool:
        return isinstance(self.__value, InstanceId)

    def as_dict(self, camel_case: bool = True) -> dict[str, T_ID]:
        if isinstance(self.__value, InstanceId):
            return {self.name(camel_case): self.__value.dump(camel_case=camel_case, include_instance_type=False)}  # type: ignore[dict-item]
        else:
            return {self.name(camel_case): self.__value}

    def as_tuple(self, camel_case: bool = True) -> tuple[str, T_ID]:
        return self.name(camel_case), self.__value


class UserIdentifier:
    def __init__(self, value: str) -> None:
        self.__value: str = value

    def name(self, camel_case: bool = False) -> str:
        return "userIdentifier" if camel_case else "user_identifier"

    def as_dict(self, camel_case: bool = True) -> dict[str, str]:
        return {self.name(camel_case): self.__value}

    def as_primitive(self) -> str:
        return self.__value


class Username:
    def __init__(self, value: str) -> None:
        self.__value: str = value

    def name(self, camel_case: bool = False) -> str:
        return "username"

    def as_dict(self, camel_case: bool = True) -> dict[str, str]:
        return {self.name(camel_case): self.__value}

    def as_primitive(self) -> str:
        return self.__value


class Tablename:
    def __init__(self, value: str) -> None:
        self.__value: str = value

    def name(self, camel_case: bool = False) -> str:
        return "tablename"

    def as_dict(self, camel_case: bool = True) -> dict[str, str]:
        return {self.name(camel_case): self.__value}

    def as_primitive(self) -> str:
        return self.__value


class LimitId:
    def __init__(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError(f"Expected limitId to be of type str, got {value} of type {type(value)}")
        self.__value: str = value

    def name(self, camel_case: bool = False) -> str:
        return "limitId" if camel_case else "limit_id"

    def as_dict(self, camel_case: bool = True) -> dict[str, str]:
        return {self.name(camel_case): self.__value}

    def as_primitive(self) -> str:
        return self.__value


class WorkflowVersionIdentifier:
    def __init__(self, version: str, workflow_external_id: str) -> None:
        self.__version: str = version
        self.__workflow_external_id: str = workflow_external_id

    def as_dict(self, camel_case: bool = True) -> dict[str, str]:
        return {
            "version": self.__version,
            ("workflowExternalId" if camel_case else "workflow_external_id"): self.__workflow_external_id,
        }

    def as_primitive(self) -> NoReturn:
        raise AttributeError(f"Not supported for {type(self).__name__} implementation")


class DataModelingIdentifier:
    def __init__(
        self,
        space: str,
        external_id: str | None = None,
        version: str | None = None,
        instance_type: Literal["node", "edge"] | None = None,
    ):
        self.__space = space
        self.__external_id = external_id
        self.__version = version
        self.__instance_type = instance_type

    def as_dict(self, camel_case: bool = True) -> dict[str, str]:
        output = {"space": self.__space}
        if self.__external_id is not None:
            key = "externalId" if camel_case else "external_id"
            output[key] = self.__external_id
        if self.__version is not None:
            output["version"] = self.__version
        if self.__instance_type is not None:
            output["instanceType"] = self.__instance_type
        return output

    def as_primitive(self) -> NoReturn:
        raise AttributeError(f"Not supported for {type(self).__name__} implementation")


class ExternalId(Identifier[str]): ...


class InternalId(Identifier[int]): ...


class PrincipalIdentifier:
    def __init__(self, id: str | None = None, external_id: str | None = None) -> None:
        if not exactly_one_is_not_none(id, external_id):
            raise ValueError("Exactly one of id or external_id must be specified, got both or neither.")
        if id is not None and not isinstance(id, str):
            raise TypeError(f"Expected id to be of type str, got {type(id)}")
        if external_id is not None and not isinstance(external_id, str):
            raise TypeError(f"Expected external_id to be of type str, got {type(external_id)}")
        self.__id = id
        self.__external_id = external_id

    def as_dict(self, camel_case: bool = True) -> dict[str, str]:
        output: dict[str, str] = {}
        if self.__id is not None:
            output["id"] = self.__id
        if self.__external_id is not None:
            output["externalId" if camel_case else "external_id"] = self.__external_id
        return output

    def as_primitive(self) -> str:
        if self.__id is not None:
            return self.__id
        if self.__external_id is not None:
            return self.__external_id
        raise ValueError("Neither id nor external_id is set, cannot return primitive value")


T_Identifier = TypeVar("T_Identifier", bound=IdentifierCore)


class IdentifierSequenceCore(Generic[T_Identifier], ABC):
    def __init__(self, identifiers: list[T_Identifier], is_singleton: bool) -> None:
        self._identifiers = identifiers
        self.__is_singleton = is_singleton

    def __len__(self) -> int:
        return len(self._identifiers)

    def __getitem__(self, item: int) -> T_Identifier:
        return self._identifiers[item]

    def is_singleton(self) -> bool:
        return self.__is_singleton

    def assert_singleton(self) -> None:
        if not self.is_singleton():
            raise ValueError("Exactly one of id or external_id (or instance_id if accepted) must be specified")

    def as_singleton(self) -> SingletonIdentifierSequence:
        self.assert_singleton()
        return cast(SingletonIdentifierSequence, self)

    def chunked(self: T_IdentifierSequenceCore, chunk_size: int) -> list[T_IdentifierSequenceCore]:
        return [
            type(self)(chunk, is_singleton=self.is_singleton())
            for chunk in split_into_chunks(self._identifiers, chunk_size)
        ]

    def as_dicts(self) -> list[dict[str, int | str]]:
        return [identifier.as_dict() for identifier in self._identifiers]

    def as_primitives(self) -> list[int | str]:
        return [identifier.as_primitive() for identifier in self._identifiers]

    def are_unique(self) -> bool:
        return len(self) == len(set(self.as_primitives()))

    @staticmethod
    def unwrap_identifier(identifier: str | int | dict) -> str | int | InstanceId:
        if isinstance(identifier, (str, int)):
            return identifier
        if "externalId" in identifier:
            return identifier["externalId"]
        if "id" in identifier:
            return identifier["id"]
        if "space" in identifier:
            return identifier["space"]
        if "instanceId" in identifier:
            return InstanceId.load(identifier["instanceId"])
        raise ValueError(f"{identifier} does not contain 'id' or 'externalId', or 'space'")

    @staticmethod
    def extract_identifiers(dct: dict[str, Any]) -> dict[str, str | int]:
        """An API payload might look like {"id": 1, "before": "2w-ago", ...}. This function extracts the identifiers"""
        return {k: dct[k] for k in ("id", "externalId", "instanceId") if k in dct}


T_IdentifierSequenceCore = TypeVar("T_IdentifierSequenceCore", bound=IdentifierSequenceCore)


class IdentifierSequence(IdentifierSequenceCore[Identifier]):
    @overload
    @classmethod
    def of(cls, *ids: list[int | str | InstanceId]) -> IdentifierSequence: ...

    @overload
    @classmethod
    def of(cls, *ids: int | str | InstanceId) -> IdentifierSequence: ...

    @classmethod
    def of(cls, *ids: int | str | InstanceId | Sequence[int | str | InstanceId]) -> IdentifierSequence:
        if len(ids) == 1 and isinstance(ids[0], Sequence) and not isinstance(ids[0], str):
            return cls([Identifier(val) for val in ids[0]], is_singleton=False)
        else:
            return cls([Identifier(val) for val in ids], is_singleton=len(ids) == 1)

    @classmethod
    def load(
        cls,
        ids: int | Sequence[int] | None = None,
        external_ids: str | Sequence[str] | SequenceNotStr[str] | None = None,
        instance_ids: InstanceId
        | Sequence[InstanceId | tuple[str, str] | dict[str, str]]
        | tuple[str, str]
        | dict[str, str]
        | None = None,
        *,
        id_name: str = "",
    ) -> IdentifierSequence:
        if id_name and not id_name.endswith("_"):
            id_name += "_"
        value_passed_as_primitive = False
        all_identifiers: list[int | str | InstanceId] = []

        if ids is not None:
            if isinstance(ids, numbers.Integral):
                value_passed_as_primitive = True
                all_identifiers.append(int(ids))
            elif isinstance(ids, Sequence) and not isinstance(ids, str):
                all_identifiers.extend([int(id_) for id_ in ids])
            else:
                raise TypeError(f"{id_name}ids must be of type int or Sequence[int]. Found {type(ids)}")

        if external_ids is not None:
            if isinstance(external_ids, str):
                value_passed_as_primitive = True
                all_identifiers.append(external_ids)
            elif isinstance(external_ids, Sequence):
                all_identifiers.extend([str(extid) for extid in external_ids])
            else:
                raise TypeError(
                    f"{id_name}external_ids must be of type str or SequenceNotStr[str]. Found {type(external_ids)}"
                )
        if instance_ids is not None:
            if isinstance(instance_ids, InstanceId | dict) or (
                isinstance(instance_ids, tuple) and len(instance_ids) == 2
            ):
                value_passed_as_primitive = True
                all_identifiers.append(InstanceId.load(instance_ids))  # type: ignore[arg-type]
            elif isinstance(instance_ids, Sequence):
                all_identifiers.extend([InstanceId.load(instance_id) for instance_id in instance_ids])  # type: ignore[arg-type]
            else:
                raise TypeError(
                    f"{id_name}instance_ids must be of type InstanceId or Sequence[InstanceId]. Found {type(instance_ids)}"
                )

        is_singleton = value_passed_as_primitive and len(all_identifiers) == 1
        return cls(identifiers=[Identifier(val) for val in all_identifiers], is_singleton=is_singleton)


class IdentifierSequenceWithInstanceId(IdentifierSequence):
    # TODO: Remove 'instanceId' above and move here instead (and update usage where appropriate...)
    #       to avoid error messages mentioning instance ID where it's not relevant, and be more
    #       helpful where needed.

    def assert_singleton(self) -> None:
        if not self.is_singleton():
            raise ValueError("Exactly one of id or external_id or instance_id must be specified")


class SingletonIdentifierSequence(IdentifierSequenceCore[Identifier]): ...


class DataModelingIdentifierSequence(IdentifierSequenceCore[DataModelingIdentifier]): ...


class UserIdentifierSequence(IdentifierSequenceCore[UserIdentifier]):
    # TODO: Inferred type from inherited methods 'as_dicts' and 'as_primitives' wrongly include 'int'
    @classmethod
    def load(cls, user_identifiers: str | SequenceNotStr[str]) -> UserIdentifierSequence:
        if isinstance(user_identifiers, str):
            return cls([UserIdentifier(user_identifiers)], is_singleton=True)

        elif isinstance(user_identifiers, Sequence):
            return cls(list(map(UserIdentifier, map(str, user_identifiers))), is_singleton=False)

        raise TypeError(f"user_identifiers must be of type str or SequenceNotStr[str]. Found {type(user_identifiers)}")

    def assert_singleton(self) -> None:
        if not self.is_singleton():
            raise ValueError("Exactly one user identifier (string) must be specified")


class UsernameSequence(IdentifierSequenceCore[Username]):
    @classmethod
    def load(cls, usernames: str | SequenceNotStr[str]) -> UsernameSequence:
        if isinstance(usernames, str):
            return cls([Username(usernames)], is_singleton=True)

        elif isinstance(usernames, Sequence):
            return cls([Username(username) for username in usernames], is_singleton=False)

        raise TypeError(f"usernames must be of type str or SequenceNotStr[str]. Found {type(usernames)}")

    def assert_singleton(self) -> None:
        if not self.is_singleton():
            raise ValueError("Exactly one username (string) must be specified")

    @staticmethod
    def unwrap_identifier(identifier: str | int | dict) -> str | int | InstanceId:
        if isinstance(identifier, str):
            return identifier
        elif isinstance(identifier, dict):
            if "username" in identifier:
                return identifier["username"]
            raise ValueError(f"{identifier} does not contain 'username'.")
        raise TypeError(f"identifier must be of type str or dict. Found {type(identifier)}")


class TablenameSequence(IdentifierSequenceCore[Tablename]):
    @classmethod
    def load(cls, tablenames: str | SequenceNotStr[str]) -> TablenameSequence:
        if isinstance(tablenames, str):
            return cls([Tablename(tablenames)], is_singleton=True)

        elif isinstance(tablenames, Sequence):
            return cls([Tablename(tablename) for tablename in tablenames], is_singleton=False)

        raise TypeError(f"tablenames must be of type str or SequenceNotStr[str]. Found {type(tablenames)}")

    def assert_singleton(self) -> None:
        if not self.is_singleton():
            raise ValueError("Exactly one tablename (string) must be specified")

    @staticmethod
    def unwrap_identifier(identifier: str | int | dict) -> str | int | InstanceId:
        if isinstance(identifier, str):
            return identifier
        elif isinstance(identifier, dict):
            if "tablename" in identifier:
                return identifier["tablename"]
            raise ValueError(f"{identifier} does not contain 'tablename'.")
        raise TypeError(f"identifier must be of type str or dict. Found {type(identifier)}")


class WorkflowVersionIdentifierSequence(IdentifierSequenceCore[WorkflowVersionIdentifier]):
    @classmethod
    def load(cls, workflow_ids: Sequence[dict]) -> WorkflowVersionIdentifierSequence:
        if len(workflow_ids) == 1 and isinstance(workflow_ids[0], dict):
            return cls(
                [
                    WorkflowVersionIdentifier(
                        version=workflow_ids[0]["version"], workflow_external_id=workflow_ids[0]["workflowExternalId"]
                    )
                ],
                is_singleton=True,
            )
        elif isinstance(workflow_ids, Sequence) and workflow_ids and isinstance(workflow_ids[0], dict):
            return cls(
                [WorkflowVersionIdentifier(entry["version"], entry["workflowExternalId"]) for entry in workflow_ids],
                is_singleton=False,
            )
        raise TypeError(f"WorkflowIdentifier must be of type str or Sequence[str]. Found {type(workflow_ids)}")

    def assert_singleton(self) -> None:
        if not self.is_singleton():
            raise ValueError("Exactly one workflow version must be specified")

    @staticmethod
    def unwrap_identifier(identifier: str | dict) -> str | tuple[str, str]:  # type: ignore[override]
        if isinstance(identifier, str):
            return identifier
        if "workflowExternalId" in identifier and "version" in identifier:
            return identifier["workflowExternalId"], identifier["version"]
        raise ValueError(f"{identifier} does not contain both 'workflowExternalId' and 'version'")


class PrincipalIdentifierSequence(IdentifierSequenceCore[PrincipalIdentifier]):
    @classmethod
    def load(
        cls,
        ids: str | SequenceNotStr[str] | None = None,
        external_ids: str | SequenceNotStr[str] | None = None,
    ) -> PrincipalIdentifierSequence:
        return cls(
            identifiers=PrincipalIdentifierSequence._load_identifiers(ids, external_ids),
            is_singleton=(isinstance(ids, str) and external_ids is None)
            or (isinstance(external_ids, str) and ids is None),
        )

    @staticmethod
    def _load_identifiers(
        ids: str | SequenceNotStr[str] | None,
        external_ids: str | SequenceNotStr[str] | None,
    ) -> list[PrincipalIdentifier]:
        identifiers: list[PrincipalIdentifier] = []
        if ids is not None:
            if isinstance(ids, str):
                identifiers.append(PrincipalIdentifier(id=ids))
            elif isinstance(ids, Sequence):
                identifiers.extend([PrincipalIdentifier(id=id_) for id_ in ids])
            else:
                raise TypeError(f"ids must be of type str or Sequence[str]. Found {type(ids)}")

        if external_ids is not None:
            if isinstance(external_ids, str):
                identifiers.append(PrincipalIdentifier(external_id=external_ids))
            elif isinstance(external_ids, Sequence):
                identifiers.extend([PrincipalIdentifier(external_id=extid) for extid in external_ids])
            else:
                raise TypeError(f"external_ids must be of type str or Sequence[str]. Found {type(external_ids)}")
        return identifiers



from __future__ import annotations

import asyncio
import functools
import gzip
import logging
import platform
from collections.abc import AsyncIterator, MutableMapping
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, NoReturn, cast
from urllib.parse import urljoin

import httpx
from typing_extensions import Self

from cognite.client._http_client import AsyncHTTPClientWithRetry, AsyncHTTPClientWithRetryConfig
from cognite.client.config import global_config
from cognite.client.exceptions import (
    CogniteAPIError,
    CogniteDuplicatedError,
    CogniteHTTPStatusError,
    CogniteNotFoundError,
    CogniteProjectAccessError,
)
from cognite.client.utils import _json_extended as _json
from cognite.client.utils._auxiliary import drop_none_values
from cognite.client.utils._text import shorten
from cognite.client.utils._url import resolve_url

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
    from cognite.client.config import ClientConfig
    from cognite.client.response import CogniteHTTPResponse


logger = logging.getLogger(__name__)


@dataclass
class FailedRequestHandler:
    message: str
    status_code: int
    missing: list[str] | None
    duplicated: list[str] | None
    x_request_id: str | None
    headers: dict[str, str] | httpx.Headers
    response_headers: dict[str, str] | httpx.Headers
    extra: dict[str, Any]
    cause: CogniteHTTPStatusError
    stream: bool

    def __post_init__(self) -> None:
        self.headers = BasicAsyncAPIClient._sanitize_headers(self.headers)
        self.response_headers = BasicAsyncAPIClient._sanitize_headers(self.response_headers)

    @classmethod
    async def from_status_error(cls, err: CogniteHTTPStatusError, stream: bool) -> Self:
        response = err.response
        error, missing, duplicated = {}, None, None

        if stream:
            await response.aread()
        try:
            error = response.json()["error"]
        except (_json.JSONDecodeError, KeyError):
            message = response.text
        else:
            match error:
                case str():
                    message = error
                case dict():
                    error.pop("code", None)  # some APIs also return status code here
                    message = error.pop("message")
                    missing = error.pop("missing", None) or None  # no empty lists wanted
                    duplicated = error.pop("duplicated", None) or None
                case _:
                    message = response.text

        return cls(
            message=message,
            status_code=response.status_code,
            missing=missing,
            duplicated=duplicated,
            x_request_id=response.headers.get("x-request-id"),
            headers=err.request.headers,
            response_headers=response.headers,
            extra=error,
            cause=err,
            stream=stream,
        )

    def log_failed_request(self, payload: dict | None = None) -> None:
        response, request = self.cause.response, self.cause.request
        extra: dict[str, Any] = {
            "payload": payload,
            "missing": self.missing,
            "duplicated": self.duplicated,
            "headers": self.headers,
            "response-headers": self.response_headers,
        }
        if not self.stream:
            extra["response-payload"] = shorten(response.text, 1_000)

        if response.history:
            for res_hist in response.history:
                logger.debug(
                    f"REDIRECT AFTER HTTP Error {res_hist.status_code} {res_hist.request.method} "
                    f"{res_hist.request.url}: {res_hist.text}"
                )
        logger.debug(
            f"HTTP Error {self.status_code} {request.method} {request.url}: {self.message}",
            extra=drop_none_values(extra),
        )

    async def raise_api_error(self, cognite_client: AsyncCogniteClient) -> NoReturn:
        cluster = cognite_client._config.cdf_cluster
        project = cognite_client._config.project

        match self.status_code, self.duplicated, self.missing:
            case 401, *_:
                await self._raise_no_project_access_error(cognite_client, cluster, project)
            case 409, list(), None:
                self._raise_api_error(CogniteDuplicatedError, cluster, project)
            case 400 | 422, None, list():
                self._raise_api_error(CogniteNotFoundError, cluster, project)
            case _:
                self._raise_api_error(CogniteAPIError, cluster, project)

    async def _raise_no_project_access_error(
        self, cognite_client: AsyncCogniteClient, cluster: str | None, project: str
    ) -> NoReturn:
        maybe_projects = await CogniteProjectAccessError._attempt_to_get_projects(cognite_client, project)
        raise CogniteProjectAccessError(
            project=project,
            x_request_id=self.x_request_id,
            maybe_projects=maybe_projects,
            cluster=cluster,
        ) from None  # we don't surface the underlying CogniteHTTPStatusError

    def _raise_api_error(self, err_type: type[CogniteAPIError], cluster: str | None, project: str) -> NoReturn:
        raise err_type(
            message=self.message,
            code=self.status_code,
            x_request_id=self.x_request_id,
            missing=self.missing,
            duplicated=self.duplicated,
            extra=self.extra,
            cluster=cluster,
            project=project,
        ) from None

    @staticmethod
    def classify_error(err: BaseException) -> Literal["failed", "unknown"]:
        if isinstance(err, CogniteAPIError) and err.code >= 500:
            return "unknown"
        return "failed"


@functools.cache
def get_user_agent() -> str:
    from cognite.client import __version__

    try:
        from httpx._client import USER_AGENT
    except ImportError:
        USER_AGENT = "python-httpx/<unknown>"

    sdk_version = f"CognitePythonSDK/{__version__}"
    python_version = (
        f"{platform.python_implementation()}/{platform.python_version()} "
        f"({platform.python_build()};{platform.python_compiler()})"
    )
    os_version_info = [platform.release(), platform.machine(), platform.architecture()[0]]
    operating_system = f"{platform.system()}/{'-'.join(s for s in os_version_info if s)}"

    return f"{USER_AGENT} {sdk_version} {python_version} {operating_system}"


class BasicAsyncAPIClient:
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        self._config = config
        self._api_version = api_version
        self._api_subversion = config.api_subversion
        self._cognite_client = cognite_client
        self._init_async_http_clients()

        self._CREATE_LIMIT = 1000
        self._LIST_LIMIT = 1000
        self._RETRIEVE_LIMIT = 1000
        self._DELETE_LIMIT = 1000
        self._UPDATE_LIMIT = 1000

    def _init_async_http_clients(self) -> None:
        self._http_client = AsyncHTTPClientWithRetry(
            config=AsyncHTTPClientWithRetryConfig(status_codes_to_retry={429}, max_retries_read=0),
            refresh_auth_header=self._refresh_auth_header,
        )
        self._http_client_with_retry = AsyncHTTPClientWithRetry(
            config=AsyncHTTPClientWithRetryConfig(),
            refresh_auth_header=self._refresh_auth_header,
        )

    def __getstate__(self) -> dict[str, Any]:
        """Prepare object for pickling by removing unpicklable async clients."""
        state = self.__dict__.copy()
        state.pop("_http_client")
        state.pop("_http_client_with_retry")
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Restore object after unpickling."""
        self.__dict__.update(state)
        self._init_async_http_clients()

    def _select_async_http_client(self, is_retryable: bool) -> AsyncHTTPClientWithRetry:
        return self._http_client_with_retry if is_retryable else self._http_client

    @property
    def _base_url_with_base_path(self) -> str:
        if self._api_version:
            return urljoin(self._config.base_url, f"/api/{self._api_version}/projects/{self._config.project}")
        return self._config.base_url

    async def _request(
        self,
        method: Literal["GET", "PUT", "HEAD"],
        /,
        full_url: str,
        content: bytes | AsyncIterator[bytes] | None = None,
        headers: dict[str, Any] | None = None,
        timeout: float | None = None,
        include_cdf_headers: bool = False,
        api_subversion: str | None = None,
    ) -> CogniteHTTPResponse:
        """
        Make a request to something that is outside Cognite Data Fusion, with retry enabled.
        Requires the caller to handle errors coming from non-2xx response status codes.

        Args:
            method (Literal['GET', 'PUT', 'HEAD']): HTTP method.
            full_url (str): Full URL to make the request to.
            content (bytes | AsyncIterator[bytes] | None): Optional body content to send along with the request.
            headers (dict[str, Any] | None): Optional headers to include in the request.
            timeout (float | None): Override the default timeout for this request.
            include_cdf_headers (bool): Whether to include Cognite Data Fusion headers in the request. Defaults to False.
            api_subversion (str | None): When include_cdf_headers=True, override the API subversion to use for the request. Has no effect otherwise.

        Returns:
            CogniteHTTPResponse: The response from the server.

        Raises:
            CogniteHTTPStatusError: If the response status code is 4xx or 5xx.
        """
        http_client = self._select_async_http_client(method in {"GET", "PUT", "HEAD"})
        if include_cdf_headers:
            headers = self._configure_headers(additional_headers=headers, api_subversion=api_subversion)
        try:
            res = await http_client.request(
                method, full_url, content=content, headers=headers, timeout=timeout or self._config.timeout
            )
            self._log_successful_request(res)
            return res

        except CogniteHTTPStatusError as err:
            handler = await FailedRequestHandler.from_status_error(err, stream=False)
            handler.log_failed_request()
            raise

    @asynccontextmanager
    async def _stream(
        self,
        method: Literal["GET", "POST"],
        /,
        *,
        url_path: str | None = None,
        full_url: str | None = None,
        json: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        full_headers: dict[str, Any] | None = None,
        timeout: float | None = None,
        api_subversion: str | None = None,
    ) -> AsyncIterator[CogniteHTTPResponse]:
        assert url_path or full_url, "Either url_path or full_url must be provided"
        full_url = full_url or resolve_url(self, method, cast(str, url_path))[1]
        if full_headers is None:
            full_headers = self._configure_headers(headers, api_subversion)

        ctx = self._http_client_with_retry.stream(
            method, full_url, json=json, headers=full_headers, timeout=timeout or self._config.timeout
        )
        try:
            async with ctx as resp:
                # Request is sent when entering the context manager, so if we get here, it was successful:
                self._log_successful_request(resp, payload=json, stream=True)
                yield resp

        except CogniteHTTPStatusError as err:
            await self._handle_status_error(err, payload=json, stream=True)

    async def _get(
        self,
        url_path: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        follow_redirects: bool = False,
        api_subversion: str | None = None,
        semaphore: asyncio.BoundedSemaphore | None = None,
    ) -> CogniteHTTPResponse:
        _, full_url = resolve_url(self, "GET", url_path)
        full_headers = self._configure_headers(additional_headers=headers, api_subversion=api_subversion)
        try:
            res = await self._http_client_with_retry.request(
                "GET",
                full_url,
                params=params,
                headers=full_headers,
                follow_redirects=follow_redirects,
                timeout=self._config.timeout,
                semaphore=semaphore,
            )
        except CogniteHTTPStatusError as err:
            await self._handle_status_error(err)

        self._log_successful_request(res)
        return res

    async def _post(
        self,
        url_path: str,
        json: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        follow_redirects: bool = False,
        api_subversion: str | None = None,
        semaphore: asyncio.BoundedSemaphore | None = None,
    ) -> CogniteHTTPResponse:
        is_retryable, full_url = resolve_url(self, "POST", url_path)
        full_headers = self._configure_headers(additional_headers=headers, api_subversion=api_subversion)
        # We want to control json dumping, so we pass it along to httpx.Client.post as 'content'
        content = self._handle_json_dump(json, full_headers)

        http_client = self._select_async_http_client(is_retryable)
        try:
            res = await http_client.request(
                "POST",
                full_url,
                content=content,
                params=params,
                headers=full_headers,
                follow_redirects=follow_redirects,
                timeout=self._config.timeout,
                semaphore=semaphore,
            )
        except CogniteHTTPStatusError as err:
            await self._handle_status_error(err, payload=json)

        self._log_successful_request(res, payload=json)
        return res

    async def _put(
        self,
        url_path: str,
        content: str | bytes | AsyncIterator[bytes] | None = None,
        json: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        follow_redirects: bool = False,
        api_subversion: str | None = None,
        timeout: float | None = None,
        semaphore: asyncio.BoundedSemaphore | None = None,
    ) -> CogniteHTTPResponse:
        _, full_url = resolve_url(self, "PUT", url_path)

        full_headers = self._configure_headers(additional_headers=headers, api_subversion=api_subversion)
        if content is None:
            content = self._handle_json_dump(json, full_headers)
        try:
            res = await self._http_client_with_retry.request(
                "PUT",
                full_url,
                content=content,
                params=params,
                headers=full_headers,
                follow_redirects=follow_redirects,
                timeout=timeout or self._config.timeout,
                semaphore=semaphore,
            )
        except CogniteHTTPStatusError as err:
            await self._handle_status_error(err, payload=json)

        self._log_successful_request(res, payload=json)
        return res

    def _configure_headers(
        self, additional_headers: dict[str, str] | None, api_subversion: str | None
    ) -> dict[str, str]:
        from cognite.client import __version__

        # We use latin-1 to mimic requests' behavior and avoid UnicodeEncodeError; httpx flat out
        # refuses non-ascii (which is correct per RFC 7230). We cast because the rest of the code
        # base expects str, not bytes, but bytes is perfectly fine
        client_name = cast(str, self._config.client_name.encode("latin-1"))
        headers = {
            "content-type": "application/json",
            "accept": "application/json",
            "x-cdp-sdk": "CognitePythonSDK:" + __version__,
            "x-cdp-app": client_name,
            "cdf-version": api_subversion or self._api_subversion,
            "user-agent": get_user_agent(),
            **self._config.headers,
            **(additional_headers or {}),
        }
        self._refresh_auth_header(headers)
        return headers

    def _refresh_auth_header(self, headers: MutableMapping[str, Any]) -> None:
        auth_header_name, auth_header_value = self._config.credentials.authorization_header()
        headers[auth_header_name] = auth_header_value

    async def _handle_status_error(
        self, error: CogniteHTTPStatusError, payload: dict[str, Any] | None = None, stream: bool = False
    ) -> NoReturn:
        """The response had an HTTP status code of 4xx or 5xx"""
        handler = await FailedRequestHandler.from_status_error(error, stream=stream)
        handler.log_failed_request(payload)
        await handler.raise_api_error(self._cognite_client)

    def _log_successful_request(
        self, res: CogniteHTTPResponse, payload: dict[str, Any] | None = None, stream: bool = False
    ) -> None:
        extra: dict[str, Any] = {
            "headers": self._sanitize_headers(res.request.headers),
            "payload": payload,
            "response-headers": dict(res.headers),
        }
        if not stream and self._config.debug:
            extra["response-payload"] = shorten(res.text, 1_000)

        logger.debug(
            f"{res.http_version} {res.request.method} {res.url} {res.status_code}",
            extra=drop_none_values(extra),
        )

    @staticmethod
    def _handle_json_dump(json: dict[str, Any] | None, full_headers: MutableMapping[str, str]) -> bytes | str | None:
        if json is None:
            return None

        content = _json.dumps_no_nan_or_inf(json)
        if global_config.disable_gzip:
            return content

        full_headers["Content-Encoding"] = "gzip"
        return gzip.compress(content.encode())

    @staticmethod
    def _sanitize_headers(headers: httpx.Headers | dict[str, str]) -> dict[str, str]:
        sanitized = dict(headers)
        for k in sanitized.keys():
            if k.lower() in {"authorization", "proxy-authorization"}:
                sanitized[k] = "***"
        return sanitized



from __future__ import annotations

from collections.abc import AsyncIterator, Sequence
from typing import Any, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import raw
from cognite.client.utils._auxiliary import split_into_chunks, unpack_items_in_payload
from cognite.client.utils._concurrency import AsyncSDKTask, execute_async_tasks
from cognite.client.utils._url import interpolate_and_url_encode
from cognite.client.utils._validation import assert_type
from cognite.client.utils.useful_types import SequenceNotStr


class RawTablesAPI(APIClient):
    _RESOURCE_PATH = "/raw/dbs/{}/tables"

    @overload
    def __call__(self, db_name: str, chunk_size: None = None) -> AsyncIterator[raw.Table]: ...

    @overload
    def __call__(self, db_name: str, chunk_size: int) -> AsyncIterator[raw.TableList]: ...

    async def __call__(
        self, db_name: str, chunk_size: int | None = None, limit: int | None = None
    ) -> AsyncIterator[raw.Table | raw.TableList]:
        """Iterate over tables

        Fetches tables as they are iterated over, so you keep a limited number of tables in memory.

        Args:
            db_name (str): Name of the database to iterate over tables for
            chunk_size (int | None): Number of tables to return in each chunk. Defaults to yielding one table a time.
            limit (int | None): Maximum number of tables to return. Defaults to return all items.

        Yields:
            raw.Table | raw.TableList: The tables in the database.
        """
        table_iterator = self._list_generator(
            list_cls=raw.TableList,
            resource_cls=raw.Table,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, db_name),
            chunk_size=chunk_size,
            method="GET",
            limit=limit,
        )
        async for item in self._set_db_name_on_tables_generator(table_iterator, db_name):
            yield item

    @overload
    async def create(self, db_name: str, name: str) -> raw.Table: ...

    @overload
    async def create(self, db_name: str, name: list[str]) -> raw.TableList: ...

    async def create(self, db_name: str, name: str | list[str]) -> raw.Table | raw.TableList:
        """`Create one or more tables. <https://developer.cognite.com/api#tag/Raw/operation/createTables>`_

        Args:
            db_name (str): Database to create the tables in.
            name (str | list[str]): A table name or list of table names to create.

        Returns:
            raw.Table | raw.TableList: raw.Table or list of tables that has been created.

        Examples:

            Create a new table in a database:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.raw.tables.create("db1", "table1")
        """
        assert_type(name, "name", [str, Sequence])
        if isinstance(name, str):
            items: dict[str, Any] | list[dict[str, Any]] = {"name": name}
        else:
            items = [{"name": n} for n in name]
        tb = await self._create_multiple(
            list_cls=raw.TableList,
            resource_cls=raw.Table,
            input_resource_cls=raw.TableWrite,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, db_name),
            items=items,
        )
        return self._set_db_name_on_tables(tb, db_name)

    async def delete(self, db_name: str, name: str | SequenceNotStr[str]) -> None:
        """`Delete one or more tables. <https://developer.cognite.com/api#tag/Raw/operation/deleteTables>`_

        Args:
            db_name (str): Database to delete tables from.
            name (str | SequenceNotStr[str]): A table name or list of table names to delete.

        Examples:

            Delete a list of tables:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.raw.tables.delete("db1", ["table1", "table2"])
        """
        assert_type(name, "name", [str, Sequence])
        if isinstance(name, str):
            name = [name]

        url_path = interpolate_and_url_encode(self._RESOURCE_PATH, db_name) + "/delete"
        tasks = [
            AsyncSDKTask(self._post, url_path=url_path, json={"items": [{"name": n} for n in chunk]})
            for chunk in split_into_chunks(name, self._DELETE_LIMIT)
        ]
        summary = await execute_async_tasks(tasks)
        summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=unpack_items_in_payload, task_list_element_unwrap_fn=lambda el: el["name"]
        )

    @overload
    def _set_db_name_on_tables(self, tb: raw.Table, db_name: str) -> raw.Table: ...

    @overload
    def _set_db_name_on_tables(self, tb: raw.TableList, db_name: str) -> raw.TableList: ...

    def _set_db_name_on_tables(self, tb: raw.Table | raw.TableList, db_name: str) -> raw.Table | raw.TableList:
        if isinstance(tb, raw.Table):
            tb._db_name = db_name
            return tb
        elif isinstance(tb, raw.TableList):
            for t in tb:
                t._db_name = db_name
            return tb
        raise TypeError("tb must be raw.Table or raw.TableList")

    async def _set_db_name_on_tables_generator(
        self, table_iterator: AsyncIterator[raw.Table | raw.TableList], db_name: str
    ) -> AsyncIterator[raw.Table | raw.TableList]:
        async for tbl in table_iterator:
            yield self._set_db_name_on_tables(tbl, db_name)

    async def list(self, db_name: str, limit: int | None = DEFAULT_LIMIT_READ) -> raw.TableList:
        """`List tables <https://developer.cognite.com/api#tag/Raw/operation/getTables>`_

        Args:
            db_name (str): The database to list tables from.
            limit (int | None): Maximum number of tables to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            raw.TableList: List of requested tables.

        Examples:

            List the first 5 tables:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> table_list = client.raw.tables.list("db1", limit=5)

            Iterate over tables, one-by-one:

                >>> for table in client.raw.tables(db_name="db1"):
                ...     table  # do something with the table

            Iterate over chunks of tables to reduce memory load:

                >>> for table_list in client.raw.tables(db_name="db1", chunk_size=25):
                ...     table_list # do something with the tables
        """
        tb = await self._list(
            list_cls=raw.TableList,
            resource_cls=raw.Table,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, db_name),
            method="GET",
            limit=limit,
        )
        return self._set_db_name_on_tables(tb, db_name)



"""
===============================================================================
6b64355411f69544440ffc5607c18fb1
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.limits import Limit, LimitList
from cognite.client.utils._async_helpers import run_sync

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
from cognite.client.data_classes.filters import Prefix


class SyncLimitsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def retrieve(self, id: str) -> Limit | None:
        """
        `Retrieve a limit value by its id. <https://api-docs.cognite.com/20230101-alpha/tag/Limits/operation/fetchLimitById/>`_

        Retrieves a limit value by its `limitId`.

        Args:
            id (str): Limit ID to retrieve.
                Limits are identified by an id containing the service name and a service-scoped limit name.
                For instance `atlas.monthly_ai_tokens` is the id of the `atlas` service limit `monthly_ai_tokens`.
                Service and limit names are always in `lower_snake_case`.

        Returns:
            Limit | None: The requested limit, or `None` if not found.

        Examples:

            Retrieve a single limit by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.limits.retrieve(id="atlas.monthly_ai_tokens")
        """
        return run_sync(self.__async_client.limits.retrieve(id=id))

    def list(self, filter: Prefix | None = None, limit: int | None = DEFAULT_LIMIT_READ) -> LimitList:
        """
        `List all limit values <https://api-docs.cognite.com/20230101-alpha/tag/Limits/operation/listLimits/>`_

        Retrieves all limit values for a specific project. Optionally filter by limit ID prefix using a `Prefix` filter.

        Args:
            filter (Prefix | None): Optional `Prefix` filter to apply on the `limitId` property (only `Prefix` filters are supported).
            limit (int | None): Maximum number of limits to return. Defaults to 25. Set to None or -1 to return all limits

        Returns:
            LimitList: List of all limit values in the project.

        Examples:

            List all limits:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> limits = client.limits.list(limit=None)

            List limits filtered by prefix (e.g., all limits for the 'atlas' service):

                >>> from cognite.client.data_classes.filters import Prefix
                >>> prefix_filter = Prefix("limitId", "atlas.")
                >>> limits = client.limits.list(filter=prefix_filter)
        """
        return run_sync(self.__async_client.limits.list(filter=filter, limit=limit))



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
)

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class Principal(CogniteResource, ABC):
    _type: ClassVar[str]

    def __init__(self, id: str) -> None:
        self.id = id

    @classmethod
    @abstractmethod
    def _load_principal(cls, resource: dict[str, Any]) -> Self:
        """Load a principal from a resource dictionary."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        type_ = resource.get("type")
        if type_ is None and hasattr(cls, "_type"):
            type_ = cls._type
        elif type_ is None:
            raise KeyError("type")
        principal_cls = _PRINCIPAL_CLS_BY_TYPE.get(type_.lower())
        if principal_cls is None:
            return UnknownPrincipal._load_principal(resource)  # type: ignore[return-value]
        return cast(
            Self,
            principal_cls._load_principal(
                resource,
            ),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the principal to a dictionary."""
        output = super().dump(camel_case=camel_case)
        output["type"] = self._type.upper()
        return output


class UserPrincipal(Principal):
    """Represents a user principal in Cognite Data Fusion (CDF).

    Arguments:
        id (str): The ID of an organization user
        name (str): Human-readable name of the principal
        picture_url (str): URL to a picture of the principal
        email (str | None): User email. Do not use this to uniquely identify a user, as it can be changed
            and is not guaranteed to be unique. Use the id field instead.
        given_name (str | None): The given name of the user
        middle_name (str | None): The middle name of the user
        family_name (str | None): The family name of the user

    """

    _type = "user"

    def __init__(
        self,
        id: str,
        name: str,
        picture_url: str,
        email: str | None = None,
        given_name: str | None = None,
        middle_name: str | None = None,
        family_name: str | None = None,
    ) -> None:
        super().__init__(id)
        self.name = name
        self.picture_url = picture_url
        self.email = email
        self.given_name = given_name
        self.middle_name = middle_name
        self.family_name = family_name

    @classmethod
    def _load_principal(cls, resource: dict[str, Any]) -> Self:
        return cls(
            id=resource["id"],
            name=resource["name"],
            picture_url=resource["pictureUrl"],
            email=resource.get("email"),
            given_name=resource.get("givenName"),
            middle_name=resource.get("middleName"),
            family_name=resource.get("familyName"),
        )


@dataclass
class ServiceAccountCreator(CogniteObject):
    """The creator of a service account.

    Arguments:
        org_id (str): The ID of an organization.
        user_id (str): The ID of an organization user

    """

    org_id: str
    user_id: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(org_id=resource["orgId"], user_id=resource["userId"])


class ServicePrincipal(Principal):
    """Represents a service account principal in Cognite Data Fusion (CDF).

    Arguments:
        id (str): Unique identifier of a service account
        name (str): Human-readable name of the service account
        created_by (ServiceAccountCreator): The creator of the service account
        created_time (int): When the principal was created. It is given as the number of milliseconds
            since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int): When the principal was last updated.  It is given as the number of milliseconds
            since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        picture_url (str): URL to a picture of the principal.
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        description (str | None): A description of the service account.

    """

    _type = "service_account"

    def __init__(
        self,
        id: str,
        name: str,
        created_by: ServiceAccountCreator,
        created_time: int,
        last_updated_time: int,
        picture_url: str,
        external_id: str | None = None,
        description: str | None = None,
    ) -> None:
        super().__init__(id)
        self.name = name
        self.created_by = created_by
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.picture_url = picture_url
        self.external_id = external_id
        self.description = description

    @classmethod
    def _load_principal(cls, resource: dict[str, Any]) -> Self:
        return cls(
            id=resource["id"],
            name=resource["name"],
            created_by=ServiceAccountCreator._load(resource["createdBy"]),
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            picture_url=resource["pictureUrl"],
            external_id=resource.get("externalId"),
            description=resource.get("description"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["createdBy" if camel_case else "created_by"] = self.created_by.dump(camel_case=camel_case)
        return output


class UnknownPrincipal(Principal):
    """Represents an unknown principal in Cognite Data Fusion (CDF).

    This class is used when the principal type is not recognized or not defined in the SDK.
    Typically, this can happen when a new type of principal is introduced in CDF that is not yet supported by the SDK.

    Arguments:
        id (str): Unique identifier of the principal.
        type (str): The type of the principal, which is not recognized by the SDK.
        data (dict[str, Any]): Additional data associated with the principal, excluding the 'id' and 'type' fields.

    """

    _type = "unknown"

    def __init__(self, id: str, type: str, data: dict[str, Any]) -> None:
        super().__init__(id)
        self.type = type
        self.__data = data

    @classmethod
    def _load_principal(cls, resource: dict[str, Any]) -> Self:
        return cls(
            id=resource["id"],
            type=resource["type"],
            data={k: v for k, v in resource.items() if k not in {"id", "type"}},
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {"id": self.id, "type": self.type, **self.__data}


class PrincipalList(CogniteResourceList[Principal]):
    _RESOURCE = Principal

    def as_ids(self) -> list[str]:
        """Returns a list of principal IDs."""
        return [principal.id for principal in self]

    def _build_id_mappings(self) -> None:
        # Override as the base implementation assumes that if the first element has an 'external_id' attribute,
        # then all elements have it. This is not the case for principals.
        return


# Build the mapping AFTER all classes are defined
_PRINCIPAL_CLS_BY_TYPE: dict[str, type[Principal]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in Principal.__subclasses__()
    if hasattr(subclass, "_type") and not getattr(subclass, "__abstractmethods__", None)
}



from __future__ import annotations

import importlib
from types import ModuleType
from typing import TypeVar, overload

_T = TypeVar("_T")


@overload
def local_import(m1: str, /) -> ModuleType: ...


@overload
def local_import(m1: str, m2: str, /) -> tuple[ModuleType, ModuleType]: ...


@overload
def local_import(m1: str, m2: str, m3: str, /) -> tuple[ModuleType, ModuleType, ModuleType]: ...


@overload
def local_import(m1: str, m2: str, m3: str, m4: str, /) -> tuple[ModuleType, ModuleType, ModuleType, ModuleType]: ...


def local_import(*module: str) -> ModuleType | tuple[ModuleType, ...]:
    from cognite.client.exceptions import CogniteImportError

    if len(module) == 1:
        name = module[0]
        try:
            return importlib.import_module(name)
        except ImportError as e:
            raise CogniteImportError(name.split(".")[0]) from e

    modules = []
    for name in module:
        try:
            modules.append(importlib.import_module(name))
        except ImportError as e:
            raise CogniteImportError(name.split(".")[0]) from e
    return tuple(modules)



from __future__ import annotations

import textwrap
from typing import Any

from cognite.client._api_client import APIClient
from cognite.client.data_classes.data_modeling import DataModelIdentifier
from cognite.client.data_classes.data_modeling.graphql import DMLApplyResult
from cognite.client.data_classes.data_modeling.ids import DataModelId
from cognite.client.exceptions import CogniteGraphQLError, GraphQLErrorSpec
from cognite.client.utils._url import interpolate_and_url_encode


class DataModelingGraphQLAPI(APIClient):
    async def _post_graphql(self, url_path: str, query_name: str, json: dict) -> dict[str, Any]:
        res = (await self._post(url_path=url_path, json=json)).json()
        # Errors can be passed both at top level and nested in the response:
        errors = res.get("errors", []) + ((res.get("data", {}).get(query_name) or {}).get("errors") or [])
        if errors:
            raise CogniteGraphQLError([GraphQLErrorSpec.load(error) for error in errors])
        return res["data"]

    async def _unsafely_wipe_and_regenerate_dml(self, id: DataModelIdentifier) -> str:
        """Wipe and regenerate the DML for a given data model.

        Note:
            This removes all comments from the DML.

        Args:
            id (DataModelIdentifier): The data model to apply DML to.

        Returns:
            str: The new DML
        """
        graphql_body = """
            query WipeAndRegenerateDml($space: String!, $externalId: String!, $version: String!) {
                unsafelyWipeAndRegenerateDmlBasedOnDataModel(space: $space, externalId: $externalId, version: $version) {
                    items {
                        graphQlDml
                    }
                }
            }
        """
        data_model_id = DataModelId.load(id)
        payload = {
            "query": textwrap.dedent(graphql_body),
            "variables": {
                "space": data_model_id.space,
                "externalId": data_model_id.external_id,
                "version": data_model_id.version,
            },
        }

        query_name = "unsafelyWipeAndRegenerateDmlBasedOnDataModel"
        res = await self._post_graphql(url_path="/dml/graphql", query_name=query_name, json=payload)
        return res[query_name]["items"][0]["graphQlDml"]

    async def apply_dml(
        self,
        id: DataModelIdentifier,
        dml: str,
        name: str | None = None,
        description: str | None = None,
        previous_version: str | None = None,
    ) -> DMLApplyResult:
        """Apply the DML for a given data model.

        Args:
            id (DataModelIdentifier): The data model to apply DML to.
            dml (str): The DML to apply.
            name (str | None): The name of the data model.
            description (str | None): The description of the data model.
            previous_version (str | None): The previous version of the data model. Specify to reuse view versions from previous data model version.

        Returns:
            DMLApplyResult: The id of the updated data model.

        Examples:

            Apply DML:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.data_modeling.graphql.apply_dml(
                ...     id=("mySpaceExternalId", "myModelExternalId", "1"),
                ...     dml="type MyType { id: String! }",
                ...     name="My model name",
                ...     description="My model description"
                ... )
        """
        graphql_body = """
            mutation UpsertGraphQlDmlVersion($dmCreate: GraphQlDmlVersionUpsert!) {
                upsertGraphQlDmlVersion(graphQlDmlVersion: $dmCreate) {
                    errors {
                        kind
                        message
                        hint
                        location {
                            start {
                                line
                                column
                            }
                        }
                    }
                    result {
                        space
                        externalId
                        version
                        name
                        description
                        graphQlDml
                        createdTime
                        lastUpdatedTime
                    }
                }
            }
        """
        data_model_id = DataModelId.load(id)
        payload = {
            "query": textwrap.dedent(graphql_body),
            "variables": {
                "dmCreate": {
                    "space": data_model_id.space,
                    "externalId": data_model_id.external_id,
                    "version": data_model_id.version,
                    "previousVersion": previous_version,
                    "graphQlDml": dml,
                    "name": name,
                    "description": description,
                }
            },
        }

        query_name = "upsertGraphQlDmlVersion"
        res = await self._post_graphql(url_path="/dml/graphql", query_name=query_name, json=payload)
        return DMLApplyResult.load(res[query_name]["result"])

    async def query(
        self, id: DataModelIdentifier, query: str, variables: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """Execute a GraphQl query against a given data model.

        Args:
            id (DataModelIdentifier): The data model to query.
            query (str): The query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.

        Returns:
            dict[str, Any]: The query result

        Examples:

            Execute a graphql query against a given data model:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.data_modeling.graphql.query(
                ...     id=("mySpace", "myDataModel", "v1"),
                ...     query="listThings { items { thingProperty } }",
                ... )
        """
        dm_id = DataModelId.load(id)
        endpoint = interpolate_and_url_encode(
            "/userapis/spaces/{}/datamodels/{}/versions/{}/graphql", dm_id.space, dm_id.external_id, dm_id.version
        )
        return await self._post_graphql(url_path=endpoint, query_name="", json={"query": query, "variables": variables})
